@@ -353,7 +353,7 @@ class KStandardAction:
         action = Action(mainWindow)
         action.triggered.connect(mainWindow.configureToolBar)
         action.setText(i18n('Configure Tool&bars...'))
-        action.setIcon(KIcon('configure-toolbars'))  # TODO: winprep
+        action.setIcon(KIcon('configure-toolbars'))
         action.setIconText(i18n('Configure toolbars'))
         separator = Action(mainWindow)
         separator.setSeparator(True)
@@ -1361,11 +1361,11 @@ class KSeparator(QFrame):
 class ToolBarItem(QListWidgetItem):
 
     """a toolbar item"""
-    emptyIcon:QIcon
 
     def __init__(self, action:Action, parent:QListWidget) ->None:
         self.action = action
         self.parent = parent
+        self.emptyIcon:QIcon
         QListWidgetItem.__init__(self, self.__icon(), self.__text(), parent)
         # drop between items, not onto items
         self.setFlags(
@@ -1375,7 +1375,7 @@ class ToolBarItem(QListWidgetItem):
         """the action icon, default is an empty icon"""
         result = self.action.icon()
         if result.isNull():
-            if not self.emptyIcon:
+            if not hasattr(self, 'emptyIcon'):
                 iconSize = self.parent.style().pixelMetric(
                     QStyle.PixelMetric.PM_SmallIconSize)
                 _ = QPixmap(iconSize, iconSize)
@@ -1508,7 +1508,7 @@ class KEditToolBar(KDialog):
         self._removeAction.setEnabled(bool(toolItem))
         if toolItem:
             self.upAction.setEnabled(bool(row))
-            self.downAction.setEnabled(row < len(self.activeList) - 1)
+            self.downAction.setEnabled(row < self.activeList.count() - 1)
         else:
             self.upAction.setEnabled(False)
             self.downAction.setEnabled(False)
@@ -1566,7 +1566,7 @@ class KEditToolBar(KDialog):
         if Internal.mainWindow is None:
             return
         activeActions = (cast(ToolBarItem, self.activeList.item(
-            x)).action for x in range(len(self.activeList)))
+            x)).action for x in range(self.activeList.count()))
         names = {
             v: k for k,
             v in Internal.mainWindow.actionCollection(
