@@ -8,19 +8,21 @@ Read the user manual for a description of the interface to this scoring engine
 """
 
 import itertools
+from typing import List, Tuple, TYPE_CHECKING, Iterable, Dict, Any
 
-from tile import Tile
-from meld import Meld, MeldList
+from tile import Tile, Meld, MeldList
 
+if TYPE_CHECKING:
+    from tile import Tiles
 
 class Permutations:
 
     """creates permutations for building melds out of single tiles.
     NEVER returns Kongs!"""
-    cache = {}
-    permuteCache = {}
+    cache : Dict[tuple, 'Permutations'] = {}
+    permuteCache : Dict[tuple, Tuple[Tuple[Tuple, ...], ...]] = {}
 
-    def __new__(cls, tiles):
+    def __new__(cls, tiles:'Tiles') ->'Permutations':
         cacheKey = tuple(x.key for x in tiles)
         if cacheKey in cls.cache:
             return cls.cache[cacheKey]
@@ -28,44 +30,46 @@ class Permutations:
         cls.cache[cacheKey] = result
         return result
 
-    def __init__(self, tiles):
+    def __init__(self, tiles:'Tiles') ->None:
         self.tiles = tiles
         if not hasattr(self, 'variants'):
             self.variants = self._variants()
 
-    def _variants(self):
+    def _variants(self) ->List[MeldList]:
         """full Meld lists"""
         honors = []
         for tile in sorted(set(self.tiles)):
             if tile.isHonor:
-                count = self.tiles.count(tile)
+                those = list(x for x in self.tiles if x == tile)
+                count = len(those)
                 if count == 4:
-                    honors.append(tile.single)
+                    honors.append(Meld([those.pop()]))
                     count -= 1
-                honors.append(tile.meld(count))
+                honors.append(Meld(those))
         boni = [x.single for x in self.tiles if x.isBonus]
-        variants = []
+        variants:List[List[MeldList]] = []
         for group in Tile.colors.upper():
             gTiles = [x for x in self.tiles if x.group == group]
-            groupVariants = self.__colorVariants(group, [x.value for x in gTiles])
+            groupVariants = self.__colorVariants(group, [x.value for x in gTiles])  # type:ignore[misc]
             if groupVariants:
                 variants.append(groupVariants)
-        result = []
+        result:List[MeldList] = []
+        variant:Any
         for variant in (sum(x, []) for x in itertools.product(*variants)):
             if variant not in result:
-                result.append(variant)
+                result.append(MeldList(variant))
         result = sorted(MeldList(honors + x + boni) for x in result)
         return result
 
     @classmethod
-    def permute(cls, valuesTuple):
+    def permute(cls, valuesTuple:Tuple[int, ...]) ->Tuple[Tuple[Tuple[int, ...], ...], ...]:
         """return all groupings into melds.
         values is a tuple of int, range 1..9"""
         assert isinstance(valuesTuple, tuple)
         if valuesTuple in cls.permuteCache:
             return cls.permuteCache[valuesTuple]
         values = list(valuesTuple)
-        result = list()
+        result:List[Tuple[Tuple, ...]] = []
         possibleMelds = []
         valueSet = set(values)
         for value in sorted(valueSet):
@@ -86,18 +90,18 @@ class Permutations:
                     for combi in permuteRest:
                         result.append(tuple(list(appendValue) + list(combi)))
                 else:
-                    result.append(appendValue)
+                    result.append(tuple(appendValue))
         else:
-            result = list([list([tuple([x]) for x in values])])  # pylint: disable=consider-using-generator
+            result = list([tuple([tuple([x]) for x in values])])  # pylint: disable=consider-using-generator
             # the solution proposed by pylint creates something unusable
         tupleResult = tuple(sorted(set(tuple(tuple(sorted(x)) for x in result))))
         cls.permuteCache[valuesTuple] = tupleResult
         return tupleResult
 
-    colorPermCache = {}
+    colorPermCache : Dict[tuple, tuple] = {}
 
     @classmethod
-    def usefulPermutations(cls, values):
+    def usefulPermutations(cls, values:Iterable) ->Tuple:
         """return all variants usable for standard MJ formt (4 melds plus 1 pair),
         and also the variant with the most pungs. At least one will be returned.
         This is meant for the standard MJ format (4 pungs/kongs/chows plus 1 pair)"""
@@ -121,7 +125,9 @@ class Permutations:
                     maxPungs = pungCount
                     maxPungVariant = variant
             if maxPungs > 0 and maxPungVariant not in result:
+                assert maxPungVariant
                 result.append(maxPungVariant)
+            assert minMeldVariant
             result.append(minMeldVariant)
             if not result:
                 # if nothing seems useful, return all possible permutations
@@ -130,7 +136,7 @@ class Permutations:
         return cls.colorPermCache[values]
 
     @classmethod
-    def __colorVariants(cls, color, values):
+    def __colorVariants(cls, color:str, values:Iterable[int]) ->List[MeldList]:
         """generates all possible meld variants out of original
         where values is a string like '113445'.
         Returns lists of Meld"""
@@ -143,9 +149,9 @@ class Permutations:
                 groups.append(content)
                 allValues = [x for x in allValues if x > border]
         combinations = [cls.usefulPermutations(x) for x in groups]
-        result = []
+        result:List[MeldList] = []
         for variant in itertools.product(*combinations):
-            melds = []
+            melds = MeldList()
             for block in variant:
                 for meld in block:
                     melds.append(Meld(Tile(color, x) for x in meld))
