@@ -2,19 +2,8 @@
 
 """Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>
 
-Kajongg is free software you can redistribute it and/or modifys
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation either version 2 of the License, or
-(at your option) any later version.
+SPDX-License-Identifier: GPL-2.0
 
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program if not, write to the Free Software
-Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
 # See the user manual for a description of how to define rulesets.
@@ -22,6 +11,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 # The KDE translation teams will "automatically" translate name and
 # description into many languages.
 
+from typing import Optional
+
 from rule import PredefinedRuleset
 from mi18n import i18nE, i18n
 
@@ -32,16 +23,18 @@ class ClassicalChinese(PredefinedRuleset):
     for local variants. This should be defined such that the
     sum of the differences to the local variants is minimized."""
 
-    def __init__(self, name=None):
+    def __init__(self, name:Optional[str]=None) ->None:
         PredefinedRuleset.__init__(
             self,
             name or i18nE('Classical Chinese standard'))
 
-    def _initRuleset(self):
-        """sets the description"""
+    def _initRuleset(self) ->None:
+        """set the description"""
+        assert isinstance(self.raw_data, str)
+        self.name = self.raw_data
         self.description = i18n('Classical Chinese')
 
-    def addManualRules(self):
+    def addManualRules(self) ->None:
         """those are actually winner rules but in the kajongg scoring mode they must be selected manually"""
         # applicable only if we have a concealed meld and a declared kong:
         self.winnerRules.createRule(
@@ -88,13 +81,13 @@ class ClassicalChinese(PredefinedRuleset):
             'East won nine times in a row', 'FEastWonNineTimesInARow||Orotate', limits=1,
             description=i18n('If that happens, East gets a limit score and the winds rotate'))
 
-    def addPenaltyRules(self):
+    def addPenaltyRules(self) ->None:
         """as the name says"""
         self.penaltyRules.createRule(
             'False Naming of Discard, Claimed for Mah Jongg and False Declaration of Mah Jongg',
             'Oabsolute payers=2 payees=2', points=-300)
 
-    def addHandRules(self):
+    def addHandRules(self) ->None:
         """as the name says"""
         self.handRules.createRule(
             'Own Flower and Own Season',
@@ -110,7 +103,7 @@ class ClassicalChinese(PredefinedRuleset):
             'Long Hand', r'FLongHand||Oabsolute', points=0,
             description=i18n('The hand contains too many tiles'))
 
-    def addParameterRules(self):
+    def addParameterRules(self) ->None:
         """as the name says"""
         self.parameterRules.createRule(
             'Points Needed for Mah Jongg',
@@ -150,8 +143,17 @@ class ClassicalChinese(PredefinedRuleset):
         self.parameterRules.createRule(
             'Standard Rotation',
             'FStandardRotation||Orotate||Ointernal')
+        self.parameterRules.createRule(
+            'Discard tiles ordered',
+            'booldiscardTilesOrdered', parameter=False,
+            description=i18n('Discard tiles ordered from top left'))
+        self.parameterRules.createRule(
+            'Leave holes for claimed discards',
+            'booldiscardTilesOrderedLeaveHole', parameter=True,
+            description=i18n(
+                'Only when discarding tiles ordered: Do not place the next discard where the claimed discard was'))
 
-    def loadRules(self):
+    def loadRules(self) ->None:
         """define the rules"""
         self.addParameterRules()  # must be first!
         self.addPenaltyRules()
@@ -305,18 +307,18 @@ class ClassicalChineseDMJL(ClassicalChinese):
 
     """classical chinese rules, German rules"""
 
-    def __init__(self, name=None):
+    def __init__(self, name:Optional[str]=None) ->None:
         ClassicalChinese.__init__(
             self,
             name or i18nE('Classical Chinese DMJL'))
 
-    def _initRuleset(self):
-        """sets the description"""
+    def _initRuleset(self) ->None:
+        """set the description"""
         ClassicalChinese._initRuleset(self)
         self.description = i18n(
             'Classical Chinese as defined by the Deutsche Mah Jongg Liga (DMJL) e.V.')
 
-    def loadRules(self):
+    def loadRules(self) ->None:
         ClassicalChinese.loadRules(self)
         # the squirming snake is only covered by standard mahjongg rule if
         # tiles are ordered
@@ -336,7 +338,7 @@ class ClassicalChineseDMJL(ClassicalChinese):
         self.handRules.createRule('Big Four Joys', 'FBigFourJoys', doubles=2,
                                   description=i18n('4 Pungs or Kongs of winds'))
 
-        self.winnerRules['OnlyHonors'].doubles = 2
+        self.winnerRules['OnlyHonors'].doubles = 2  # type:ignore[attr-defined]
 
         self.penaltyRules.createRule(
             'False Naming of Discard, Claimed for Chow',
@@ -362,26 +364,26 @@ class ClassicalChineseBMJA(ClassicalChinese):
 
     """classical chinese rules, British rules"""
 
-    def __init__(self, name=None):
+    def __init__(self, name:Optional[str]=None) ->None:
         ClassicalChinese.__init__(
             self,
             name or i18nE('Classical Chinese BMJA'))
 
-    def _initRuleset(self):
-        """sets the description"""
+    def _initRuleset(self) ->None:
+        """set the description"""
         ClassicalChinese._initRuleset(self)
         self.description = i18n(
             'Classical Chinese as defined by the British Mah-Jong Association')
 
-    def addParameterRules(self):
+    def addParameterRules(self) ->None:
         """those differ for BMJA from standard"""
         ClassicalChinese.addParameterRules(self)
-        self.parameterRules['kongBoxSize'].parameter = 14
-        self.parameterRules['maxChows'].parameter = 1
-        self.parameterRules['limit'].parameter = 1000
-        self.parameterRules['mustDeclareCallingHand'].parameter = True
+        self.parameterRules['kongBoxSize'].parameter = 14  # type:ignore[attr-defined]
+        self.parameterRules['maxChows'].parameter = 1  # type:ignore[attr-defined]
+        self.parameterRules['limit'].parameter = 1000  # type:ignore[attr-defined]
+        self.parameterRules['mustDeclareCallingHand'].parameter = True  # type:ignore[attr-defined]
 
-    def loadRules(self):
+    def loadRules(self) ->None:
         ClassicalChinese.loadRules(self)
         del self.winnerRules['ZeroPointHand']
         originalCall = self.winnerRules.pop('MahJonggwithOriginalCall')
@@ -497,7 +499,7 @@ class ClassicalChineseBMJA(ClassicalChinese):
             'FCallingHand||Ohand=Purity',
             doubles=3)
 
-def load():
+def load() ->None:
     """load predefined rulesets.
     add new predefined rulesets here.
     """
